/*
 * huffman.cpp
 *
 *  Created on: Jan 21, 2009
 *      Author: jonathanmasci
 */

#include "huffman.h"
#include <math.h>

int huffman::_symbol_length = 0;

huffman::huffman(char *input, int input_length, int symbol_length) :
  _input(input),
  _input_length(input_length),
  _symbols_count(0)
{
  _symbol_length = symbol_length;
}

huffman::~huffman()
{
}

void
huffman::encode()
{
  computeWeights();
  buildTree();
  buildCodeTable();

}

void
huffman::decode() { }

double
huffman::getEntropy()
{
  std::map<char*,double>::iterator it;
  double entropy = 0;

  for (it = _frequencies.begin(); it != _frequencies.end(); ++it)
      entropy += (*it).second * log2((*it).second);

  return -entropy;
}

double
huffman::getWeightedPathLength()
{
  std::map<char *, std::pair<char *, int>, _map_compare>::iterator it;
  float weighted_path = 0;

  for (it = _codewords.begin(); it != _codewords.end(); ++it)
    weighted_path += (*it).second.second * _frequencies[(*it).first];

  return weighted_path;
}

/**
* Compute the probability of each symbol and store it in a map structure.
* The key is a char* and the comparing is done by memcmp.
*/
void
huffman::computeWeights()
{
  int i;
  int j;
  int symbols_count = 0;
  std::map<char*,double>::iterator it;
  double prob_sum = 0;
  char *buff;

  for (i = 0; i < _input_length; i += _symbol_length)
    {
      buff = new char[_symbol_length];
      for (j = 0; j < _symbol_length && (j + i) < _input_length; ++j)
        buff[j] = _input[j + i];
      symbols_count++;
      _frequencies[buff]++;
    }
  _symbols_count = symbols_count;

  // Compute the relative probability and put nodes in the queue
  for (it = _frequencies.begin(); it != _frequencies.end(); ++it)
    {
      (*it).second /= symbols_count;
      prob_sum += (*it).second;
      // This is unnecessary. Just to have probabilities stored for the entropy.
      _freqs_queue.push(std::pair<char *, double>((*it).first, (*it).second));
      // Insert in the nodes queue
      Node *n = new Node();
      n->symbol = (*it).first;
      n->freq = (*it).second;
      n->parent = n->left = n->right = NULL;
      _nodes_queue.push(n);
    }
  // assert(prob_sum == 1.0);
}

/**
* Public method to build the codes table.
*/
void
huffman::buildCodeTable()
{
  std::vector<char> *code = new std::vector<char>();
  buildCodeTable(_root, code);
  delete code;
}

/**
* Private method to build the codes table.
*
* @param n the root node
* @param code the code written until the current node
*/
void
huffman::buildCodeTable(Node *n, std::vector<char> *code)
{
  if (n->left == NULL && n->right == NULL) // if is leaf
    {
      char *s = (char *)malloc(sizeof(*s) * code->size());
      memset(s, 0, code->size());
      size_t i = 0;
      while (i < code->size())
        {
          s[i] = (*code)[i];
          ++i;
        }
      s[code->size()] = '\0';

      _codewords[n->symbol] =
        std::pair<char *, int>(s, code->size());
    }
  else
    {
      if (n->left != NULL)
        {
          code->push_back('0');
          buildCodeTable(n->left, code);
          code->pop_back();
        }
      if (n->right != NULL)
        {
          code->push_back('1');
          buildCodeTable(n->right, code);
          code->pop_back();
        }
    }
}

/**
* Build the codewords tree using the heap in O(n log n).
*
* @return the pointer to the tree root node
*/
void
huffman::buildTree()
{
  Node *inner, *left, *right;

  while(_nodes_queue.size() > 1)
    {
      // Get the two nodes with lowest probability.
      left = _nodes_queue.top();
      _nodes_queue.pop();
      right = _nodes_queue.top();
      _nodes_queue.pop();

      // Create a new inner node to attach the previous nodes.
      inner = (Node *)malloc(sizeof(*inner));
      inner->parent = NULL;
      inner->left = left;
      inner->right = right;
      inner->freq = left->freq + right->freq;
      left->parent = right->parent = inner;

      // Push back the newly created node into the priority queue.
      _nodes_queue.push(inner);
    }

  // Remains only one node and it is our root node
  inner = _nodes_queue.top();
  _nodes_queue.pop();

  _root = inner;
}

/**
* Return the estimated size of the compressed in bits.
*
* @return bit size estimation of the input
*/
int
huffman::getBitSizeEstimation()
{
  int i;
  int j;
  int bits = 0;

  for (i = 0; i < _input_length; i += _symbol_length)
    {
      char *buff = new char[_symbol_length];
      for (j = 0; j < _symbol_length && (j + i) < _input_length; ++j)
        buff[j] = _input[j + i];
      bits += _codewords[buff].second;
      delete[] buff;
    }
  return bits;
}

/**
* Return the size of the codewords dictionary.
*
* @return dictionary size in bits.
*/
int
huffman::getDictionarySize()
{
  int dict_size = 0;
  std::map<char *, std::pair<char *, int>, _map_compare>::iterator it;

  for (it = _codewords.begin(); it != _codewords.end(); ++it)
    dict_size += (*it).second.second;

  return dict_size;
}

/**
* Print out the weigth computed.
*/
void
huffman::printWeights()
{
  std::map<char*,double>::iterator it;
  for(it = _frequencies.begin(); it != _frequencies.end(); ++it)
    printf("%s: %f\n", (*it).first, (*it).second);
}

/**
* Print out the codewords.
*/
void
huffman::printCodeTable()
{
  std::map<char*, std::pair<char *, int> >::iterator it;
  for(it = _codewords.begin(); it != _codewords.end(); ++it)
    printf("%s: %s (%i)\n", (*it).first, ((*it).second).first,
        ((*it).second).second);
}
