/**
 * @file huffman.h
 *
 * @date Jan 21, 2009
 * @author: Jonathan Masci
 */

#ifndef HUFFMAN_H_
#define HUFFMAN_H_

#include <string>
#include <stdio.h>
#include <map>
#include <queue>
#include <vector>

class huffman
{
public:
  huffman(char *input, int input_length, int symbol_length);
  virtual ~huffman();
  void encode();
  void decode();
  double getEntropy();
  double getWeightedPathLength();
  void computeWeights();
  void printWeights();
  void buildTree();
  void buildCodeTable();
  void printCodeTable();
  int getBitSizeEstimation();
  int getDictionarySize();
  int getSymbolsCount() {return _symbols_count;};
private:
  /**
   * Return true if the probability of second element is bigger than the
   * probability of the first one.
   *
   * Doing this to keep smallest on top because priority_queue keeps biggest
   * on top.
   */
  struct _freq_compare
  {
    bool operator() (const std::pair<char *, double>& lhs,
                     const std::pair<char *, double>& rhs) const
    {
      return lhs.second > rhs.second;
    }
  };
  /**
   * Lexicographically sorted keys
   */
  struct _map_compare
  {
    bool operator() (const char *lhs, const char *rhs) const
    {
      return memcmp(lhs, rhs, _symbol_length) < 0;
    }
  };
  /**
   * The node of the binary tree
   */
  typedef struct node
  {
    struct node *parent;
    struct node *left;
    struct node *right;
    double freq;
    char *symbol;
  } Node;
  /**
   * Node comparing operator used to keep node sorted by frequency.
   */
  struct _node_compare
  {
    bool operator() (const struct node *lhs, const struct node *rhs) const
    {
      return (lhs->freq > rhs->freq);
    }
  };

  std::map<char *, double, _map_compare> _frequencies;
  std::priority_queue<Node *, std::vector<Node *>,											_node_compare> _nodes_queue;
    std::priority_queue<std::pair<char *, double>,
    std::vector<std::pair<char *, double> >,
    _freq_compare> _freqs_queue;
  Node *_root;
  char *_input;
  static int _symbol_length;
  int _input_length;
  int _symbols_count;
  std::map<char *, std::pair<char *, int>, _map_compare> _codewords;

  void buildCodeTable(Node *n, std::vector<char> *code);
};
#endif /* HUFFMAN_H_ */
